import numpy as np
from .polytools import companion
from . import matrix


class rebaser:
    def __init__(self, P, u):
        n = P.degree()
        M = companion(P)
        VV = np.stack([np.dot(np.linalg.matrix_power(M, i), u) for i in range(n)])
        self.mat = matrix.inv(VV)
        assert self.mat is not None, f"bad vector {u} with {P} :\n{VV}"

    def topoly(self, v):
        return np.dot(self.mat, v)



from .polytools import companion
import numpy as np


def revector(fa, lcfa, P):
    "try to associate a single vector to every state of fa"
    ok = True
    n = P.degree()
    M = companion(P)
    vecmap = {}
    mq0 = fa.getinitial()
    lq0 = lcfa.getinitial()
    v0 = tuple([0] * n)
    z = (mq0, lq0, v0)
    seen = set()
    todo = [z]
    seen.add(z)
    while todo:
        (ms, ls, vs) = todo.pop()
        if ms in vecmap and vs not in vecmap[ms]:
            ok = False
        daset = vecmap.get(ms, set())
        daset.add(vs)
        vecmap[ms] = daset
        tl = {}
        tm = {}
        for a, v, lt in lcfa.gettrans(ls):
            tl[a] = (tuple(np.dot(M, vs) + v), lt)
        for a, _, mt in fa.gettrans(ms):
            tm[a] = mt
        for a in tm:
            (vt, lt) = tl[a]
            mt = tm[a]
            z = (mt, lt, vt)
            if z not in seen:
                seen.add(z)
                todo.append(z)
    if ok:
        print(f"Re-vectorization ok! ({len(vecmap)} states)")
    else:
        print("#### Re-vectorization KO")
        for q, s in vecmap.items():
            if len(s) > 1:
                print(f"{q:6d} : {s}")



from argparse import HelpFormatter
import textwrap


def wrap_paragraphs(text: str, width: int, indent: str):
    """
    https://stackoverflow.com/a/75344754
    Wrapper around `textwrap.wrap()` which keeps newlines in the input string
    intact.
    """
    lines = list[str]()

    for i in text.splitlines():
        paragraph_lines = textwrap.wrap(
            i, width, initial_indent=indent, subsequent_indent=indent
        )

        # `textwrap.wrap()` will return an empty list when passed an empty
        # string (which happens when there are two consecutive line breaks in
        # the input string). This would lead to those line breaks being
        # collapsed into a single line break, effectively removing empty lines
        # from the input. Thus, we add an empty line in that case.
        lines.extend(paragraph_lines or [""])

    return lines


class Formatter(HelpFormatter):
    def _split_lines(self, text, width):
        return wrap_paragraphs(text, width, "")

    def _fill_text(self, text, width, indent):
        return "\n".join(wrap_paragraphs(text, width, indent))



from .dfau import dfa
from .arraysorter import arraysorter, indexer
from copy import deepcopy


def todfa(q0s, n, S, A, T, F, mapa):
    S = deepcopy(S)
    A = deepcopy(A)
    T = deepcopy(T)
    m = len(S)
    arraysorter([S, A, T])
    sindex = indexer(S, n)
    sindex.append(m - 1)
    seen = set()
    cur = tuple(sorted(list(q0s)))
    seen.add(cur)
    todo = [cur]
    res = dfa()
    res.setinitial(cur)
    while todo:
        cur = todo.pop()
        h = {}
        for s in cur:
            if s in F:
                res.setfinal(cur)
            i = sindex[s]
            if i >= 0:
                while i < m and S[i] == s:
                    a = A[i]
                    ah = h.get(a, set())
                    ah.add(T[i])
                    h[a] = ah
                    i += 1
        for a in h:
            nxt = tuple(sorted(list(h[a])))
            res.addtrans(cur, mapa(a), None, nxt)
            if nxt not in seen:
                seen.add(nxt)
                todo.append(nxt)
    return res


def rev(aa, follow=None):
    F = [aa.q0]
    if follow:
        todo = [aa.q0]
        seen = set()
        seen.add(aa.q0)
        while todo:
            cur = todo.pop()
            for a, _, t in aa.gettransi(cur):
                if a == follow and t not in seen:
                    todo.add(t)
                    seen.add(t)
                    F.append(t)
    return todfa(aa.F, aa.statecount(), aa.T, aa.A, aa.S, set(F), aa.ma.rev)



import numpy as np
from .polytools import commonpoly
from .dfau import dfa
from .intmap import mapper

ALPHABET = "abcdefghijklmnopqrstuvwxyz"


def str2sub(s):
    res = {}
    if "->" in s:
        for l in s.strip().split(","):
            (a, u) = l.strip().split("->")
            res[a.strip()] = u.strip()
    else:
        for i, u in enumerate(s.strip().split("/")):
            res[str(i)] = u.strip()
    return res


def sub2str(h):
    return ", ".join(f"{a}->{sa}" for a, sa in sorted(h.items()))


def canon(s, a, alpha=ALPHABET):
    h = str2sub(s)
    res = []
    ms = mapper()
    seen = set()
    todo = [a]
    seen.add(a)
    while todo:
        cur = todo.pop()
        s = f"{ alpha[ms[cur]] }->"
        for x in h[cur]:
            s += alpha[ms[x]]
            if x not in seen:
                seen.add(x)
                todo.append(x)
        res.append(s)
    return ", ".join(sorted(res))


def alpha(s):
    return sorted(list(s.keys()))


def alphaim(s):
    return sorted(list(set(sum(map(list, s.values()), []))))


def seqalpha(s):
    k = 0
    for _, v in s.items():
        k = max(k, len(v))
    return list(range(k))


def substmat(s):
    "compute the adjacency matrix of substitution [s]"
    alph = alpha(s)
    return np.array([[s[i].count(j) for j in alph] for i in alph])


def substpoly(s):
    return np.polynomial.Polynomial(np.poly(substmat(s))[::-1])


def seq(s, iw=None):
    """compute the [2k] first terms of the length sequences for every letters of substitution [s]
    starting from initial weight values [iw]"""
    alph = alpha(s)
    A = substmat(s)
    k = 2 * len(alph)
    vv = []
    if iw is None:
        curv = np.array([[1]] * len(alph))
    else:
        curv = np.array([[i] for i in iw])
    for _ in range(k):
        vv.append(curv)
        curv = np.dot(A, curv)
    res = {x: [] for x in alph}
    res[""] = [0] * k
    for v in vv:
        for x, k in zip(alph, v):
            res[x].append(k[0])
    for x in res:
        res[x] = np.array(res[x])
    return res


def combiseq(bs, u):
    "compute the first terms of word [u] for a dictionnary of sequences [bs]"
    if u == "":
        return bs[""]
    return list(sum(bs[x] for x in u))


def addpoly(s, iw=None):
    "compute the minimal reccurence polynomial for addition of [s]"
    vv = seq(s, iw)
    pp = []
    for x in s:
        u = s[x]
        for i in range(len(u) - 1):
            v = u[: i + 1]
            pp.append(combiseq(vv, v))
    return commonpoly(*pp)


def substdfa(s, a, k=0, iw=None):
    """compute the sequence value automaton for [s] from [a] with vectors of size [k]
    starting with inital weight values [iw]"""
    assert (
        s[a][0] == a
    ), f'the image of "{a}" does not starts with "{a}" in the substitution'
    res = dfa()
    vv = seq(s, iw)
    for x in s:
        u = s[x]
        for i in range(len(u)):
            v = combiseq(vv, u[:i])[:k]
            assert (
                len(v) == k
            ), f"oups... something strange with sequences lengths ({k}, {v})"
            y = u[i]
            res.addtrans(x, i, tuple(v), y)
    res.setinitial(a)
    for x in s:
        res.setfinal(x)
    return res


def substparentdfa(s, a, k=0):
    "compute the parent sequence value automaton for [s] from [a] with vectors of size [k]"
    cla = substdfa(s, a, k)
    res = dfa()
    oq0 = cla.getinitial()
    q0 = (oq0, (oq0, 0))
    res.setinitial(q0)
    seen = set()
    todo = [q0]
    seen.add(q0)
    while todo:
        ns = todo.pop()
        (s, orig) = ns
        if cla.isfinal(s):
            res.setfinal(ns)
        for a, v, t in cla.gettrans(s):
            nt = (t, (s, a))
            res.addtrans(ns, a, v, nt)
            if nt not in seen:
                seen.add(nt)
                todo.append(nt)
    return res


def dfasubst(aa):
    "compute the substitution of a prefix-closed automaton [aa]"
    h = {}
    alpha = ALPHABET
    ms = mapper()
    ms[aa.getinitial()]
    for ss, a, _, tt in aa.alltrans():
        s = alpha[ms[ss]]
        t = alpha[ms[tt]]
        if not aa.isfinal(ss):
            return None
        if type(a) is not int:
            return None
        ls = h.get(s, [])
        ls.append((a, t))
        h[s] = ls
    for s in h:
        ls = sorted(h[s])
        fs = ""
        for i, (j, c) in enumerate(ls):
            if i != j:
                return None
            fs += c
        h[s] = fs
    return h


def block(a, h, n=2):
    "compute the n-block substitution"
    if n == 1:
        return (a, h)
    if n > 2:
        return block(*block(a, h, n - 1))
    na = tuple(h[a][:2])
    todo = [na]
    name = {na: 0}
    idx = 1
    nh = dict()
    while todo:
        cur = todo.pop()
        i = name[cur]
        x = cur[0]
        m = len(h[x])
        s = []
        for x in cur:
            s.extend(list(h[x]))
        l = ""
        for k in range(m):
            nxt = tuple(s[k : k + 2])
            if nxt in name:
                j = name[nxt]
            else:
                name[nxt] = idx
                j = idx
                idx += 1
                todo.append(nxt)
            l += chr(97 + j)
        nh[chr(97 + i)] = l
    return (chr(97 + name[na]), nh)



from .polytools import companion
from .dfau import dfa
from .intmap import mapper
import numpy as np
from array import array
from . import addcut


def flatten(add, P, cut, verbose):
    n = P.degree()
    M = companion(P).astype(np.intc)
    if verbose:
        print("Expanding DFA...", flush=True, end="")
    add.ssort()
    VF = array("i", [1 if i in add.F else 0 for i in range(add.statecount())])
    VL = np.array([np.array(list(v), dtype=np.intc) for v in add.mv])
    if len(cut) == 1 and n > 1:
        cut = cut * n
    acut = np.array(cut, dtype=np.intc)
    v0 = np.array([0] * n, dtype=np.intc)
    (rn, rq0, RS, RA, RV, RT, RF) = addcut.flatten(
        add.q0, add.S, add.A, add.V, add.T, VF, add.sindex, VL, M, v0, acut, 0
    )
    res = dfa()
    res.q0 = rq0
    res.ms = mapper()
    for i in range(rn):
        res.ms[i]
    res.ma = add.ma.copy()
    res.mv = add.mv.copy()
    res.F = set(RF)
    res.S = RS
    res.A = RA
    res.V = RV
    res.T = RT
    res.ssorted = False
    res.sindex = array("i")
    if verbose:
        print("done.")
    return res



"""Precise (but slow) operations on integer matrices"""

import numpy as np
from fractions import Fraction
from .bareiss import adjugate


def inv(M):
    s, d, A = adjugate(M)
    if A is None:
        return None
    R = Fraction(s, d) * np.array(A)
    return R



"""A convenient lightweight library to construct and combine sequence automata."""

import numpy as np
from .misc import poly2str, str2poly
from .subst import (
    str2sub,
    sub2str,
    addpoly,
    substdfa,
    substparentdfa,
    dfasubst,
    substpoly,
    alpha,
    alphaim,
)
from .subst import block as doblock
from .polytools import (
    gen_pisot_root,
    companion,
    x as poly_x,
    one as poly_on,
    one as poly_one,
    naivelcm,
)
from .polybound import compcut, realcut
from .flatadd import flatten
from .dfau import prod, dfa
from pathlib import Path as P

verbose = False
stats = False
destdir = "/tmp/out"


def setparams(v, s, d):
    global verbose, stats, destdir
    verbose = v
    stats = s
    destdir = P(d)


def writefile(name, data, dir=P("Command Files")):
    genparentdir(destdir / dir / P(f"{name}.txt")).write_text(data)


def genparentdir(fname):
    parent = fname.parent
    if not parent.is_dir():
        parent.mkdir(parents=True)
    return fname


def prstats(title, fa):
    print(
        f"{title}: {fa.statecount()} states, {fa.finalcount()} final states, {fa.transcount()} transitions."
    )


def unpoly(e):
    if all(x >= 0 for x in e):
        return parry(*e)
    n = len(e)
    d = [0] * n
    for m in range(n):
        for i in range(m - 1):
            d[i] = e[i]
        d[m - 1] = e[m - 1] - 1
        for i in range(n - m):
            d[m + i] = d[i] + e[m + i]
        if all(x >= 0 for x in d):
            return parry(*d[: n - m]).tail(*d[n - m :])
    return None


class poly:
    def __init__(self, s):
        self.P = str2poly(s)

    def __repr__(self):
        return f"poly({poly2str(self.P)})"

    def image(self, s):
        return morphic(self, s)

    def subst(self):
        return unpoly([-int(x) for x in list(self.P)[::-1][1:]]).subst()


class parry:
    def __init__(self, *coeffs):
        self.coeffs = coeffs

    def __repr__(self):
        return f"parry({str(self.coeffs)[1:-1]})"

    def tail(self, *coeffs):
        return parrytail(self.coeffs, coeffs)

    def image(self, s):
        return morphic(self, s)

    def subst(self):
        h = dict()
        m = len(self.coeffs) - 1
        for i, x in enumerate(self.coeffs):
            h[str(i)] = "0" * x + (str(i + 1) if i < m else "")
        return ("0", h)


class parrytail:
    def __init__(self, u, v):
        self.u = u
        self.v = v

    def __repr__(self):
        return f"parry({str(self.u)[1:-1]}[{str(self.v)[1:-1]}]*)"

    def image(self, s):
        return morphic(self, s)

    def subst(self):
        h = dict()
        coeffs = self.u + self.v
        k = len(self.u)
        m = len(coeffs) - 1
        for i, x in enumerate(coeffs):
            h[str(i)] = "0" * x + (str(i + 1) if i < m else str(k))
        return ("0", h)


class subst:
    def __init__(self, s):
        self.a = s.strip()[0]
        self.h = str2sub(s)

    def __repr__(self):
        return f"subst({sub2str(self.h)})"

    def image(self, s):
        return morphic(self, s)

    def subst(self):
        return (self.a, self.h)


class morphic:
    def __init__(self, base, s):
        self.base = base
        self.h = str2sub(s)

    def __repr__(self):
        return f"morphic({repr(self.base)} | {sub2str(self.h)})"

    def image(self, s):
        f = str2sub(s)
        h = dict()
        for x, u in self.h.items():
            h[x] = "".join([f[y] for y in u])
        return morphic(self.base, sub2str(h))

    def subst(self):
        return self.base.subst()

    def morphism(self):
        return self.h


class block:
    def __init__(self, base, n=2):
        self.base = base
        self.n = n

    def __repr__(self):
        return f"block({repr(self.base)}, {self.n})"

    def image(self, s):
        return morphic(self, s)

    def subst(self):
        return doblock(*self.base.subst(), self.n)


class address:
    def __init__(self, base, nsname, **weights):
        self.base = base
        self.ns = nsname
        self.w = weights
        if verbose:
            print(f">>> Address automaton for {repr(self.base)}")
        self.a, self.fh = self.base.subst()
        alph = alpha(self.fh)
        morphic = hasattr(self.base, "morphism")
        self.gh = self.base.morphism() if morphic else None
        self.alpha = alphaim(self.gh) if morphic else alph
        self.P = poly_one
        iw = None
        self.rw = {x: 1 for x in self.alpha}
        if morphic:
            iw = [len(self.gh[x]) for x in alph]
            if max(iw) > 1:
                self.P *= poly_x
        if weights:
            rw = dict()
            dw = weights.get("_", 1)
            for x in self.alpha:
                rw[x] = weights.get(x, dw)
            self.rw = rw
            if morphic:
                iw = [sum([rw[y] for y in self.gh[x]]) for x in alph]
            else:
                iw = [rw[x] for x in alph]
        self.iw = iw
        self.P *= addpoly(self.fh, iw)
        self.n = self.P.degree()
        if verbose:
            print(f"  DT polynomial: {poly2str(self.P)} (n={self.n})")
        self.r = gen_pisot_root(self.P)
        if self.r is None and verbose:
            print("  *** WARNING *** the substitution is not (X^k-)Pisot!")
        elif self.r is not None and verbose:
            kX, th = self.r
            print(f"  θ={th}" + (f" (X^{kX})" if kX > 0 else ""))
        self.fa = self.gen_fa(self.n)
        if stats:
            prstats("  DT automaton", self.fa)
        if verbose:
            print()

    def gen_fa(self, n):
        if n < self.n:
            raise ValueError("Shorter vectors than base!")
        fa = substdfa(self.fh, self.a, n, self.iw)
        if self.gh is not None:
            newfa = dfa()
            alen = [0] * len(fa.ma)
            for s, a, v, t in fa.alltrans(True):
                k = len(self.gh[t])
                alen[a] = max(alen[a], k)
            aidx = []
            cur = 0
            for x in alen:
                aidx.append(cur)
                cur += x
            for i in range(sum(alen)):
                newfa.ma[i]
            newfa.setinitial(fa.getinitial())
            zv = [0] * (self.n - 1)
            for s, a, v, t in fa.alltrans(True):
                u = self.gh[t]
                k = len(u)
                if k > 0:
                    newfa.setfinal(t)
                ai = aidx[a]
                newfa.addtrans(s, ai, v, t)
                iv = v[0]
                for i, y in list(enumerate(u))[1:]:
                    di = sum(self.rw[y] for y in u[:i])
                    newfa.addtrans(s, ai + i, tuple([iv + di] + zv), y)
                    newfa.setfinal(y)
            fa = newfa
        return fa

    def state_output(self, x):
        if self.gh is None:
            return x
        if x in self.fh:
            return self.gh[x][0]
        return x

    def __repr__(self):
        return f"address({repr(self.base)}, '{self.ns}'{', ' + str(self.w)[1:-1] if self.w else ''})"

    def poly(self):
        return self.P

    def gen_ns(self, format="Walnut", minimize=True, cut=None):
        if verbose:
            print(f">>> {repr(self)} is generating it's NS")
        fa = self.fa
        if minimize:
            fa = fa.minimize()
        if stats:
            prstats("  Numeration system automaton", fa)
        fname = genparentdir(destdir / P("Custom Bases") / P(f"msd_{self.ns}.txt"))
        with open(fname, "w") as f:
            print(f"  Writing {fname} in format {format}")
            if format == "Walnut":
                fa.toWalnut(f)
            elif format == "dot":
                fa.todot(f, vformat=None)
            else:
                raise ValueError(f"Invalid format name '{format}'")
        (self + self - self).gen_dfa(
            f"msd_{self.ns}_addition",
            cut=cut,
            dir=P("Custom Bases"),
            set_ns=False,
            preverb="  ",
        )
        if verbose:
            print()

    def gen_word_automaton(self, name=None, format="Walnut"):
        if name is None:
            name = self.ns.title()
        if verbose:
            print(
                f">>> {repr(self)} is generating it's word automaton under name {name}"
            )
        fa = self.fa
        fname = genparentdir(
            destdir / P("Word Automata Library") / P(f"{name.title()}.txt")
        )
        with open(fname, "w") as f:
            if verbose:
                print(f"  Writing {fname} in format {format}\n")
            if format == "Walnut":
                fa.toWalnut(
                    f,
                    alpha=f"msd_{self.ns}",
                    outmap=lambda x: self.alpha.index(self.state_output(x)),
                )
            elif format == "dot":
                fa.todot(f, vformat=None)
            else:
                raise ValueError(f"Invalid format name '{format}'")
        if verbose:
            print()

    def __mul__(self, other):
        assert isinstance(other, int)
        return lincomb((other, self))

    def __rmul__(self, other):
        assert isinstance(other, int)
        return lincomb((other, self))

    def __add__(self, other):
        return lincomb((1, self)) + other

    def __radd__(self, other):
        return other + lincomb((1, self))

    def __sub__(self, other):
        return lincomb((1, self)) - other

    def __rsub__(self, other):
        return other - lincomb((1, self))


def opposite(xs):
    return [(-v, y) for (v, y) in xs]


class lincomb:
    def __init__(self, *xs):
        self.xs = xs

    def __add__(self, other):
        if type(other) == lincomb:
            return lincomb(*self.xs, *other.xs)
        else:
            return lincomb(*self.xs, (1, other))

    def __radd__(self, other):
        if type(other) == lincomb:
            return lincomb(*other.xs, *self.xs)
        else:
            return lincomb((1, other), *self.xs)

    def __sub__(self, other):
        if type(other) == lincomb:
            return lincomb(*self.xs, *opposite(other.xs))
        else:
            return lincomb(*self.xs, (-1, other))

    def __rsub__(self, other):
        if type(other) == lincomb:
            return lincomb(*other.xs, *opposite(self.xs))
        else:
            return lincomb((1, other), *opposite(self.xs))

    def __repr__(self):
        return "".join(
            map(
                lambda z: f"{'+' if z[0]>=0 else '-'}{(str(abs(z[0]))+'*') if abs(z[0])!=1 else ''}{repr(z[1])}",
                self.xs,
            )
        )

    def compute_fa(self):
        self.compute_poly()
        nc = len(self.xs)
        coef, addas = zip(*self.xs)
        fas = map(lambda x: x.gen_fa(self.n), addas)
        alpha = " ".join(map(lambda x: f"msd_{x.ns}", addas))

        def aux(*p):
            nonlocal nc, coef
            ra = tuple(p[:nc])
            rv = tuple(sum(ci * np.array(xi) for ci, xi in zip(coef, p[nc:])))
            return (ra, rv)

        return (prod(aux, *fas), alpha)

    def compute_poly(self):
        self.P = poly_one
        _, addas = zip(*self.xs)
        for adda in addas:
            self.P = naivelcm(self.P, adda.poly())
        self.n = self.P.degree()
        if verbose:
            print(f"  Combination polynomial is {poly2str(self.P)}")
        self.r = gen_pisot_root(self.P)
        if self.r is None and verbose:
            print("  *** WARNING *** the substitution is not (X^k-)Pisot!")
        elif self.r is not None and verbose:
            kX, th = self.r
            print(f"  θ={th}" + (f" (X^{kX})" if kX > 0 else ""))

    def gen_dfa(
        self,
        name,
        cut=None,
        vobj=0,
        vector=None,
        minimize=True,
        dir=P("Automata Library"),
        set_ns=True,
        format="Walnut",
        preverb=">>> ",
        debug=None,
    ):
        if verbose:
            print(f"{preverb}{repr(self)} is generating it's dfa under name {name}")
        fname = genparentdir(destdir / dir / P(f"{name}.txt"))
        lcfa, alpha = self.compute_fa()
        if debug:
            with open(genparentdir(destdir / P("debug") / P(f"{name}.dot")), "w") as f:
                lcfa.todot(f)
        if vector is None:
            vector = [0] * (self.n - 1) + [1]
        if stats:
            prstats("  Combination automaton", lcfa)
        if cut is None:
            if self.r is None:
                raise ValueError("the substitution is not (X^k-)Pisot!")
            cut = compcut(lcfa, self.P, vector, 0, verbose, preverb="  ")
        elif verbose:
            print(f"  Bypassing bound computation! cut={cut}")
        fa = flatten(lcfa, self.P, cut, verbose)
        if stats:
            prstats(f"  Flattened automaton ", fa)
        fa = fa.trim()
        if stats:
            prstats(f"  Trimmed flattened automaton ", fa)
        if verbose:
            obscut = realcut(fa, companion(self.P))
            print(f"  Real bound is {obscut} (vs {cut}).")
        if minimize:
            fa = fa.minimize()
            if stats:
                prstats("  After minimization", fa)
        with open(fname, "w") as f:
            if verbose:
                print(f"  Writing {fname} in format {format}\n")
            if format == "Walnut":
                fa.toWalnut(f, alpha if set_ns else None)
            elif format == "dot":
                fa.todot(f, vformat=None)
            else:
                raise ValueError(f"Invalid format name '{format}'")



#cython: language_level=3, boundscheck=False, wraparound=False
"""DFA minimization

   Cython version of the program DFA_minimizer.cc 
   Retrieved from http://users.jyu.fi/~ava/ on 22 November 2023
   Quick and dirty conversion by N. Ollinger, please read
   Copyright and conditions of use below.

   The original program header reads as follows:

/*
  Copyright Antti Valmari 2012, this commented version 2013.
  This program is from

    Antti Valmari: Fast brief practical DFA minimization,
    Information Processing Letters 112 (2012) 213-217

  You may use and adapt the program for scientific purposes at your own risk,
  but you must give credit to the original author and source. Please negotiate
  with me about uses for other purposes.

  If you do not have access to the above-mentioned publication, please see
  A. Valmari, P. Lehtinen: Efficient minimization of DFAs with partial
  transition functions, Symposium on Theoretical Aspects of Computer Science,
  2008, pp. 645-656, http://drops.dagstuhl.de/volltexte/2008/1328/
  That publication explains part of the background. However, this program is
  much further optimized.

  This program inputs a deterministic finite automaton whose transition
  function is not necessarily full, and outputs the minimal automaton
  accepting the same language. The program also contains the removal of
  irrelevant parts of the DFA.

  This program runs in O(n + m log m) time, where n is the number of states
  and m is the number of defined transitions. If the transitions are given in
  the input such that all transitions with the same label are given together,
  then the transitions with another label, and so on, then the lines
  "#include <algorithm>" and "std::sort( C.E, C.E+mm, cmp );" can be removed,
  improving the running time to O(n + m log n). These should be compared to
  the running time of Hopcroft's algorithm, which is O(nk log n), where k is
  the size of the alphabet.

  This program is also fast in practice, I believe.
*/
"""
__all__ = ["minimize"]

from libc.stdlib cimport qsort
from cpython cimport array
import array
import cython

cdef array.array iat = array.array('i', [])
cdef int w
cdef int nn
cdef int mm
cdef int[:] gM
cdef int[:] gW
cdef int[:] gA
cdef int[:] gF
cdef B
cdef C

cdef class partition:
    cdef public int z
    cdef public int[:] E
    cdef public int[:] L
    cdef public int[:] S
    cdef public int[:] F
    cdef public int[:] P

    def __init__(self, int n):
        z = 1 if n>0 else 0
        cdef int[:] E = array.clone(iat, n, zero=True)
        cdef int[:] L = array.clone(iat, n, zero=True)
        cdef int[:] S = array.clone(iat, n, zero=True)
        cdef int[:] F = array.clone(iat, n, zero=True)
        cdef int[:] P = array.clone(iat, n, zero=True)
        for i in range(n):
            E[i] = i
            L[i] = i
            S[i] = 0
        if n > 0:
            F[0] = 0
            P[0] = n
        self.z = z
        self.E = E
        self.L = L
        self.S = S
        self.F = F
        self.P = P

    def mark(self, int e):
        global w, gM, gW
        cdef int[:] E = self.E
        cdef int[:] L = self.L
        cdef int[:] S = self.S
        cdef int[:] F = self.F
        cdef int[:] P = self.P
        cdef int[:] M = gM
        cdef int[:] W = gW
        cdef int s = S[e]
        cdef int i = L[e]
        cdef int j = F[s]+M[s]
        E[i] = E[j]
        L[E[i]] = i
        E[j] = e
        L[e] = j
        if M[s]==0:
            W[w]=s
            w+=1
        M[s]+=1

    def split(self):
        global w, gM, gW
        cdef int[:] E = self.E
        cdef int[:] L = self.L
        cdef int[:] S = self.S
        cdef int[:] F = self.F
        cdef int[:] P = self.P
        cdef int[:] M = gM
        cdef int[:] W = gW
        cdef int z = self.z
        cdef int s = 0
        cdef int j = 0
        cdef int i
        while w>0:
            w-= 1
            s = W[w]
            j = F[s] + M[s]
            if j == P[s]:
                M[s] = 0
                continue
            if M[s] <= P[s]-j:
                F[z] = F[s]
                P[z] = j
                F[s] = j
            else:
                P[z] = P[s]
                F[z] = j
                P[s] = j
            for i in range(F[z],P[z]):
                S[E[i]] = z
            M[s] = 0
            M[z] = 0
            z += 1
        self.z = z


cdef make_adjacent(int[:] K):
    global nn, mm, gF, gA
    cdef int[:] F = gF
    cdef int[:] A = gA
    cdef int q
    cdef int t
    for q in range(nn+1):
        F[q]=0
    for t in range(mm):
        F[K[t]]+=1
    for q in range(nn):
        F[q+1] += F[q]
    for t in range(mm-1,-1,-1):
        F[K[t]] -= 1
        A[F[K[t]]] = t


cdef int rr

cdef reach(int q):
    global B, rr
    cdef int[:] BE = B.E
    cdef int[:] BL = B.L
    cdef int i = BL[q]
    if i >= rr:
        BE[i] = BE[rr]
        BL[BE[i]] = i
        BE[rr] = q
        BL[q] = rr
        rr += 1

cdef rem_unreachable(int[:] T, int[:] L, int[:] H):
    global B, rr, gF, gA, mm
    cdef int[:] BE = B.E
    cdef int[:] BL = B.L
    cdef int[:] BP = B.P
    cdef int[:] F = gF
    cdef int[:] A = gA
    cdef int j = 0
    make_adjacent(T)
    cdef int i=0
    while i<rr:
        for j in range(F[BE[i]], F[BE[i]+1]):
            reach(H[A[j]])
        i+=1
    j = 0
    for t in range(mm):
        if BL[T[t]] < rr:
            H[j] = H[t]
            L[j] = L[t]
            T[j] = T[t]
            j += 1
    mm = j
    BP[0] = rr
    rr = 0

cdef int *cmpL
cdef int cmp_fun(const void *i, const void *j) noexcept nogil:
    cdef int x = cmpL[(<int*>i)[0]]
    cdef int y = cmpL[(<int*>j)[0]]
    if x < y:
        return -1
    if x == y:
        return 0
    return 1

@cython.binding(True)
def minimize(int n, int m, int q0, int[:] T,  int[:] L, int[:] H,  int[:] FF, sorted=False):
    """Minimize the deterministic finite automaton given as input.

The automaton has [n] states and [m] transitions. States and transitions 
labels are encoded as integers. The initial state is [q0]. The accepting
states are listed in the array [FF]. The transitions are listed in the
arrays [T], [L], [H] of length [m].

A transition [(s,a,t) = (T[i],L[i],H[i])] reads as
"from state [s] reading symbol [a] the automaton enters state [t]".

If the transitions are given in an order such that every transitions
with the same label are given together, setting [sorted] to [True] 
avoids a useless sort and improves the overall complexity 
from O(n + m log m) to O(n + m log n).

All arrays are expected to be provided as objects that exposes writable
buffers of integers. It includes:
  - CPython [array.array] of type code ['i'] as [array('i', [1, 2, 3])]
  - NumPy arrays of dtype [np.intc] as [np.array([1, 2, 3], dtype=np.intc)]

The return value is a tuple [(n, m, q0, T, L, H ,FF)] describing the
output minimal automaton. Every array is encoded as an [array.array] of
type code ['i'].

```
from dfamin import minimize
from array import array

print(
    minimize(
        2,
        4,
        0,
        array("i", [0, 0, 1, 1]),
        array("i", [1, 0, 1, 0]),
        array("i", [0, 1, 0, 1]),
        array("i", [0, 1]),
    )
)
```
    """
    global gM, gW, gF, gA, B, C, nn, mm, rr, w, cmpL
    gL = L
    w = 0
    rr = 0
    nn = n
    mm = m
    cdef int[:] A = array.clone(iat, mm, zero=True)
    cdef int[:] F = array.clone(iat, nn + 1, zero=True)
    gA = A
    gF = F
    B = partition(nn)
    cdef int[:] BE = B.E
    cdef int[:] BL = B.L
    cdef int[:] BP = B.P
    cdef int[:] BS = B.S
    cdef int[:] BF = B.F
    cdef int q
    cdef int i
    # Remove states that cannot be reached from the initial state, and from which final states cannot be reached
    reach(q0)
    rem_unreachable(T, L, H)
    for q in FF:
        if BL[q] < BP[0]:
            reach(q)
    ff = rr
    rem_unreachable(H, L, T)
    # Make initial partition
    gM = array.clone(iat, mm + 1, zero=True)
    gW = array.clone(iat, mm + 1, zero=True)
    cdef int[:] M = gM
    cdef int[:] W = gW
    M[0] = ff
    if ff > 0:
        W[w] = 0
        w += 1
        B.split()
    # Make transition partition
    C = partition(mm)
    cdef int[:] CE = C.E
    cdef int[:] CP = C.P
    cdef int[:] CF = C.F
    cdef int[:] CS = C.S
    cdef int[:] CL = C.L
    cdef int z
    cdef int a
    cdef int t
    if mm > 0:
        if not sorted:
            cmpL = &L[0]
            qsort(&CE[0], mm, sizeof(int), &cmp_fun)
        z = 0
        M[0] = 0
        a = L[CE[0]]
        for i in range(mm):
            t = CE[i]
            if L[t] != a:
                a = L[t]
                CP[z] = i
                z += 1
                CF[z] = i
                M[z] = 0
            CS[t] = z
            CL[t] = i
        CP[z] = mm
        C.z = z + 1
    # Split blocks and cords
    make_adjacent(H)
    cdef int b = 1
    cdef int c = 0
    cdef int j
    while c < C.z:
        for i in range(CF[c], CP[c]):
            B.mark(T[CE[i]])
        B.split()
        c += 1
        while b < B.z:
            for i in range(BF[b], BP[b]):
                for j in range(F[BE[i]], F[BE[i] + 1]):
                    C.mark(A[j])
            C.split()
            b += 1
    # Count the numbers of transitions and final states in the result
    cdef int mo = 0
    cdef int fo = 0
    for t in range(mm):
        if BL[T[t]] == BF[BS[T[t]]]:
            mo += 1
    z = B.z
    for b in range(z):
        if BF[b] < ff:
            fo += 1
    resn = B.z
    resm = mo
    resq0 = BS[q0]
    resff = fo
    cdef array.array resT = array.clone(iat, resm, zero=True)
    cdef array.array resL = array.clone(iat, resm, zero=True)
    cdef array.array resH = array.clone(iat, resm, zero=True)
    cdef array.array resFF = array.clone(iat, resff, zero=True)
    cdef int cur = 0
    for t in range(mm):
        if BL[T[t]] == BF[BS[T[t]]]:
            resT[cur]=BS[T[t]]
            resL[cur]=L[t]
            resH[cur]=BS[H[t]]
            cur += 1
    cur = 0
    z = B.z
    for b in range(z):
        if BF[b] < ff:
            resFF[cur] = b
            cur += 1
    return (resn, resm, resq0, resT, resL, resH, resFF)



from array import array
from .arraysorter import arraysorter, indexer
from .intmap import mapper, swapper
from . import dfamin
from copy import deepcopy
from itertools import product
import numpy as np


def show(n, m, q0, T, L, H, FF):
    print(f"{n} {m} {q0} {len(FF)}")
    for i in range(m):
        print(f"{T[i]} {L[i]} {H[i]}")
    print(" ".join(str(i) for i in FF))


class dfa:
    def __init__(self):
        self.q0 = None
        self.ms = mapper()
        self.ma = mapper()
        self.mv = mapper()
        self.F = set()
        self.S = array("i")
        self.A = array("i")
        self.V = array("i")
        self.T = array("i")
        self.ssorted = True
        self.sindex = array("i")

    def statecount(self):
        "Count states"
        return len(self.ms)

    def transcount(self):
        "Count transitions"
        return len(self.S)

    def finalcount(self):
        return len(self.F)

    def setinitial(self, s):
        "Set initial state"
        self.q0 = self.ms[s]
        self.ssorted = False

    def getinitial(self):
        return self.ms.rev(self.q0)

    def setfinal(self, s):
        "Set state [s] as final"
        self.F.add(self.ms[s])
        self.ssorted = False

    def isfinal(self, s):
        return self.ms.get(s) in self.F

    def getfinals(self):
        for idxs in self.F:
            yield self.ms.rev(idxs)

    def addtrans(self, s, a, v, t):
        "Add a transition [(s,a,v,t)] -- NO COLLISION CHECK"
        self.S.append(self.ms[s])
        self.A.append(self.ma[a])
        self.V.append(self.mv[v])
        self.T.append(self.ms[t])
        self.ssorted = False

    def ssort(self):
        "Sort transitions by initial state"
        if self.ssorted:
            return
        arraysorter([self.S, self.A, self.V, self.T])
        self.sindex = indexer(self.S, self.statecount())
        self.ssorted = True

    def copy(self):
        res = dfa()
        res.q0 = self.q0
        res.ms = self.ms.copy()
        res.ma = self.ma.copy()
        res.mv = self.mv.copy()
        res.F = self.F.copy()
        res.S = deepcopy(self.S)
        res.A = deepcopy(self.A)
        res.V = deepcopy(self.V)
        res.T = deepcopy(self.T)
        res.ssorted = self.ssorted
        res.sindex = deepcopy(self.sindex)
        return res

    def alltransi(self):
        "Generate indexes of transition (s,a,v,t)"
        for i in range(len(self.S)):
            yield (self.S[i], self.A[i], self.V[i], self.T[i])

    def alltrans(self, sorted=False):
        "Generate all transitions"
        if sorted:
            self.ssort()
        for s, a, v, t in self.alltransi():
            yield (self.ms.rev(s), self.ma.rev(a), self.mv.rev(v), self.ms.rev(t))

    def gettransi(self, idxs):
        "Generate all transitions from state of index [idxs]"
        self.ssort()
        m = self.transcount()
        i = self.sindex[idxs]
        if i >= 0:
            while i < m and self.S[i] == idxs:
                yield (self.A[i], self.V[i], self.T[i])
                i += 1

    def gettrans(self, s):
        "Generate all transitions from states [s]"
        idxs = self.ms.get(s)
        for a, v, t in self.gettransi(idxs):
            yield (self.ma.rev(a), self.mv.rev(v), self.ms.rev(t))

    def delta(self, s, v0, red, *l):
        self.ssort()
        m = self.transcount()
        idxs = self.ms.get(s)
        cur = idxs
        vcur = v0
        for a in l:
            idxa = self.ma.get(a)
            i = self.sindex[cur]
            if i < 0:
                return None
            while i < m and self.S[i] == idxs and self.A[i] != idxa:
                i += 1
            if i == m or self.A[i] != idxa:
                return None
            v = self.mv.rev(self.V[i])
            cur = self.T[i]
            vcur = red(vcur, v)
        return vcur, cur

    def maptrans(self, f):
        "Remap [a] and [v] from every transition [(s,a,v,t)] using [(b,w)=f(a,v)] -- NO COLLISION CHECK"
        res = dfa()
        res.q0 = self.q0
        res.ms = self.ms.copy()
        res.ma = mapper()
        res.mv = mapper()
        res.F = self.F.copy()
        res.S = array("i")
        res.A = array("i")
        res.V = array("i")
        res.T = array("i")
        res.ssorted = False
        res.sindex = array("i")
        for s, a, v, t in self.alltrans():
            b, w = f(a, v)
            res.addtrans(s, b, w, t)
        return res

    def trim(self):
        "Trim automaton"
        res = dfa()
        seen = set()
        res.setinitial(self.ms.rev(self.q0))
        back = {}
        todo = [self.q0]
        seen.add(self.q0)
        while todo:
            si = todo.pop()
            for ai, vi, ti in self.gettransi(si):
                lti = back.get(ti, [])
                lti.append((si, ai, vi))
                back[ti] = lti
                if ti not in seen:
                    seen.add(ti)
                    todo.append(ti)
        coacc = set()
        todo = []
        for si in self.F:
            if si in back:
                s = self.ms.rev(si)
                res.setfinal(s)
                coacc.add(si)
                todo.append(si)
        while todo:
            ti = todo.pop()
            t = self.ms.rev(ti)
            for si, ai, vi in back[ti]:
                s = self.ms.rev(si)
                a = self.ma.rev(ai)
                v = self.mv.rev(vi)
                res.addtrans(s, a, v, t)
                if si not in coacc:
                    coacc.add(si)
                    todo.append(si)
        return res

    def minimize(self):
        "Minimize [a] loosing the vectors (copy of labels from now)"
        n = self.statecount()
        m = self.transcount()
        res = dfa()
        (n, m, q0, T, L, H, FF) = dfamin.minimize(
            n,
            m,
            self.q0,
            deepcopy(self.S),
            deepcopy(self.A),
            deepcopy(self.T),
            array("i", sorted(list(self.F))),
        )
        res.q0 = q0
        res.ms = mapper()
        for i in range(n):
            res.ms[i]
        res.ma = self.ma.copy()
        res.mv = self.ma.copy()
        res.F = set(FF)
        res.S = T
        res.A = L
        res.V = deepcopy(L)
        res.T = H
        res.ssorted = False
        res.sindex = array("i")
        return res

    def todot(self, f, vformat=str, vsep=":"):
        """Write dot representation into [f] object using [vformat] to
        format vector component. If [vformat] is [None] or return
        [None], the vector component is ignored, otherwise it
        appears on the edge separated from the label by [vsep]"""
        f.write("""digraph G {\nrankdir=LR\nnode [shape=circle]\n"" [shape=none]\n""")
        for sidx in self.F:
            f.write(f'"{self.ms.rev(sidx)}" [shape=doublecircle]\n')
        f.write(f'"" -> "{self.ms.rev(self.q0)}"\n')
        h = {}
        for s, a, v, t in self.alltrans():
            r = vformat(v) if vformat is not None else None
            if r is None:
                lbl = str(a)
            else:
                lbl = f"{a}{vsep}{r}"
            ll = h.get((s, t), [])
            ll.append(lbl)
            h[(s, t)] = ll
        nl = "\n"
        for (s, t), lbl in h.items():
            f.write(f'"{s}" -> "{t}" [label="{nl.join(lbl)}"]\n')
        f.write("}\n")

    def inferalpha(self):
        "Attempt to reconstruct DFA alphabet from transitions."
        v = next(iter(self.ma))
        if type(v) is not tuple:
            return [sorted(list(self.ma))]
        n = len(v)
        s = [set() for _ in range(n)]
        for x in self.ma:
            for i, v in enumerate(x):
                s[i].add(v)
        return [sorted(list(x)) for x in s]

    def toWalnut(self, f, alpha=None, outmap=None):
        """Write Walnut representation into [f] object using
        only [(s,a,t)] labels. If [outmap] is given, use
        it to determine output of DFAO. Otherwise use
        1 for accepting states. Alphabets are given by
        [alpha] string if given. Otherwise alphabets are
        inferred from labels."""
        self.ssort()
        swp = swapper(0, self.q0)
        if alpha is None:
            alpha = " ".join(
                "{" + ", ".join(map(str, x)) + "}" for x in self.inferalpha()
            )
        f.write(f"{alpha}\n")
        for idxs in range(self.statecount()):
            ridxs = swp[idxs]
            if outmap is None:
                v = 1 if ridxs in self.F else 0
            else:
                v = outmap(self.ms.rev(ridxs))
            f.write(f"\n{idxs} {v}\n")
            for idxa, _, ridxt in self.gettransi(ridxs):
                a = self.ma.rev(idxa)
                idxt = swp[ridxt]
                f.write(
                    f"{' '.join(map(str,a)) if type(a) is tuple else str(a)} -> {idxt}\n"
                )


def prod(f, *aa):
    res = dfa()
    res.setinitial(tuple(map(lambda a: a.getinitial(), aa)))
    for x in set(product(*map(lambda a: a.getfinals(), aa))):
        res.setfinal(x)
    for tl in product(*map(lambda a: a.alltrans(), aa)):
        sl = tuple(map(lambda x: x[0], tl))
        al = tuple(map(lambda x: x[1], tl))
        vl = tuple(map(lambda x: x[2], tl))
        tl = tuple(map(lambda x: x[3], tl))
        b, w = f(*al, *vl)
        res.addtrans(sl, b, w, tl)
    return res


def lincomb(a, coef, objv):
    "linear combinations of automaton [a] : \sum_i coef[i] * x[i] = objv"
    nc = len(coef)

    def aux(*p):
        nonlocal nc, coef
        ra = tuple(p[:nc])
        rv = tuple(sum(ci * np.array(xi) for ci, xi in zip(coef, p[nc:])))
        return (ra, rv)

    return prod(aux, *([a] * len(coef)))


def addition(a):
    return lincomb(a, [1, 1, -1], 0)



#cython: language_level=3, boundscheck=False, wraparound=False
__all__ = ["flatten"]

import cython
from cpython cimport array
import array
import numpy as np
cimport numpy as cnp

cnp.import_array()

ctypedef cnp.int32_t DTYPE_t

cdef dict msh
cdef int curs
cdef list back

cdef ms(int s, int[:] v):
    global msh, curs, back
    cdef tuple tv = (s, tuple(v))
    cdef int rv
    if tv in msh:
        return msh[tv]
    msh[tv] = curs
    rv = curs 
    back.append((s, v))
    curs += 1
    return rv

@cython.binding(True)
def flatten(int q0, int[:] S, int[:] A, int[:] V, int[:] T, int[:] VF, int[:] Sindex,  cnp.ndarray[DTYPE_t, ndim=2] VL, cnp.ndarray[DTYPE_t, ndim=2] M, cnp.ndarray[DTYPE_t, ndim=1] v0, int[:] cut, int vobj):
    global msh, curs, back
    cdef int ssidx
    cdef int ttidx
    cdef int s
    cdef int[:] vs
    cdef int a 
    cdef cnp.ndarray[DTYPE_t, ndim=1] v
    cdef int t
    cdef cnp.ndarray[DTYPE_t, ndim=1] vt
    cdef int i
    cdef array.array RS = array.array('i')
    cdef array.array RA = array.array('i')
    cdef array.array RV = array.array('i')
    cdef array.array RT = array.array('i')
    cdef array.array RF = array.array('i')
    cdef int m = len(S)
    cdef array.array todo = array.array('i')
    cdef set seen = set()
    msh = dict()
    curs = 0
    back = []
    cdef int rq0 = ms(q0, v0)
    todo.append(rq0)
    seen.add(rq0)
    while todo:
        ssidx = todo.pop()
        (s, vs) = back[ssidx]
        if VF[s] == 1 and vs[0] == vobj:
            RF.append(ssidx)
        i = Sindex[s]
        if i > -1:
            while i < m and S[i] == s:
                a = A[i]
                v = VL[V[i]]
                t = T[i]
                vt = np.dot(M, vs) + v
                if all(abs(vt) <= cut):
                    ttidx = ms(t, vt)
                    RS.append(ssidx)
                    RA.append(a)
                    RV.append(V[i])
                    RT.append(ttidx)
                    if not ttidx in seen:
                        todo.append(ttidx)
                        seen.add(ttidx)
                i += 1
    return (curs, rq0, RS, RA, RV, RT, RF)



import numpy as np
from numpy.polynomial import Polynomial as Poly
from fractions import Fraction


def PolyFrac(*a):
    "construct fraction polynomial from integer coefficients list [a]"
    return Poly(list(map(lambda x: Fraction(x), a))).trim()


def unfrac(P):
    "approximate fraction coefficients of [P] by floats"
    return Poly(list(map(float, P.coef)))


one = PolyFrac(1)
zero = PolyFrac(0)
x = PolyFrac(0, 1)


def monic(P):
    "Monic poly from [P]"
    return (Fraction(1, 1) / P.coef[-1]) * P


def bm(a):
    """Use the Berlekamp-Massey algorithm to find the minimal polynomial
    of a linearly recurrence sequence of integers a.
    As described in https://arxiv.org/pdf/2211.11721.pdf p. E4
    """
    assert len(a) % 2 == 0, "sequence length has to be even"
    n = len(a) // 2
    R1, R0 = PolyFrac(*a[::-1]), x ** (2 * n)
    if R1 == zero:
        return one
    V1, V0 = one, zero
    while R1.degree() >= n:
        R0, Q, R1 = R1, *divmod(R0, R1)
        V0, V1 = V1, V0 - Q * V1
    return monic(V1)


def naivegcd(P, Q):
    "Compute monic GCD of P and Q"
    a, b = P, Q
    while b.degree() > 0 or b.coef[0] != 0:
        a, b = b, a % monic(b)
    return monic(a)


def naivelcm(P, Q):
    "Compute LCM of P and Q"
    return P * (Q // naivegcd(P, Q))


def commonpoly(*aa):
    "Compute LCM poly of [aa] sequences"
    cur = bm(aa[0])
    for a in aa[1:]:
        nxt = bm(a)
        cur = naivelcm(cur, nxt)
    return cur


def pisot_root(P, prec=1e-4):
    "Compute the Pisot root of [P] if any"
    r = list(unfrac(P).roots())
    r.sort(key=lambda v: abs(v), reverse=True)
    if r[0].real > 1 - prec and abs(r[0].imag) < prec:
        for o in r[1:]:
            if abs(o) >= 1 - prec or abs(o) == 0:
                return None
        return r[0].real
    return None


def gen_pisot_root(P, prec=1e-4):
    "Compute the Pisot root of [P/X^k] if any"
    r = list(unfrac(P).roots())
    r.sort(key=lambda v: abs(v), reverse=True)
    k = 0
    while abs(r[-1]) < prec:
        k += 1
        r.pop(-1)
    if r[0].real > 1 - prec and abs(r[0].imag) < prec:
        for o in r[1:]:
            if abs(o) >= 1 - prec or abs(o) == 0:
                return None
        return k, r[0].real
    return None


def companion(P):
    "Compute companion matrix of integer coefficents [P]"
    d = []
    n = P.degree()
    for x in -P.coef[:-1]:
        assert x.denominator == 1, "not an integer polynomial"
        d.append(x.numerator)
    m = [[0] * i + [1] + [0] * (n - i - 1) for i in range(1, n)]
    return np.array(m + [d])



from .subst import (
    str2sub,
    sub2str,
    addpoly,
    substdfa,
    substparentdfa,
    dfasubst,
    substpoly,
    alpha,
)
from .polytools import gen_pisot_root, companion
from .polybound import compcut, realcut
from .flatadd import flatten
from .dfau import lincomb
from .misc import poly2str
from .intmap import mapper
from .lrseq import rebaser
from .mincheck import revector
from . import nfa
from numpy.polynomial import Polynomial as Poly


def prstats(title, fa):
    print(
        f"{title}: {fa.statecount()} states, {fa.finalcount()} final states, {fa.transcount()} transitions."
    )


def img(a, h, n):
    cur = a
    while len(cur) < n:
        nxt = ""
        for c in cur:
            nxt += h[c]
        cur = nxt
    return cur[:n]


class Study:
    def __init__(self, s, verbose=False, stats=False, simplify=True):
        "An API to study a given textual substitution s"
        self.verbose = verbose
        self.stats = stats
        if verbose:
            print(f"Studying substitution {s}")
        self.a = s.strip()[0]
        self.h = str2sub(s)
        self.origh = self.h
        self.origa = self.a
        if verbose:
            print(img(self.a, self.h, 70) + "...")
            print(f"Substitution polynomial: {poly2str(substpoly(self.h))}")
        if simplify:
            self.h = dfasubst(substdfa(self.h, self.a).minimize())
            self.a = "a"
            if verbose:
                print(f"Minimized substitution {sub2str(self.h)}")
                print(f"Substitution polynomial: {poly2str(substpoly(self.h))}")
        self.P = addpoly(self.h)
        self.n = self.P.degree()
        if verbose:
            print(f"Addition polynomial: {poly2str(self.P)} (n={self.n})")
        self.r = gen_pisot_root(self.P)
        if self.r is None and verbose:
            print("*** WARNING *** the substitution is not (X^k-)Pisot!")
        elif self.r is not None and verbose:
            kX, th = self.r
            print(f"θ={th}" + (f" (X^{kX})" if kX > 0 else ""))
        self.fa = substdfa(self.h, self.a, self.n)
        if self.stats:
            prstats("Substitution automaton", self.fa)

    def gendfao(self, fname, format):
        fa = substdfa(self.origh, self.origa)
        with open(fname, "w") as f:
            print(f">>> Writing {fname} in format {format}\n")
            if format == "Walnut":
                fa.toWalnut(f, outmap=lambda x: alpha(self.origh).index(x))
            elif format == "dot":
                fa.todot(f, vformat=None)
            else:
                raise ValueError(f"Invalid format name '{format}'")

    def genparentdfao(self, fname, format):
        fa = substparentdfa(self.origh, self.origa)
        out = mapper()
        with open(fname, "w") as f:
            print(f">>> Writing {fname} in format {format}\n")
            if format == "Walnut":
                fa.toWalnut(f, outmap=lambda x: out[x[1]])
            elif format == "dot":
                fa.todot(f, vformat=None)
            else:
                raise ValueError(f"Invalid format name '{format}'")

    def gennumsys(self, fname, msd, format, minimize):
        fa = self.fa
        if not msd:
            fa = nfa.rev(fa)
        if minimize:
            fa = fa.minimize()
        if self.stats:
            prstats("Numeration system automaton", fa)
        with open(fname, "w") as f:
            print(f">>> Writing {fname} in format {format}\n")
            if format == "Walnut":
                fa.toWalnut(f)
            elif format == "dot":
                fa.todot(f, vformat=None)
            else:
                raise ValueError(f"Invalid format name '{format}'")

    def genlinear(
        self,
        fname,
        lineq,
        msd,
        format,
        minimize,
        seq,
        poly,
        vector,
        cut,
        dorevector=False,
    ):
        if vector is None:
            vector = [0] * (self.n - 1) + [1]
        lcfa = lincomb(self.fa, *lineq)
        if self.stats:
            prstats("Combination automaton", lcfa)
        if seq is not None:
            with open(seq, "w") as f:
                print(f">>> Writing {seq} in format dot\n")
                lcfa.todot(f)
        if poly is not None:
            reb = rebaser(self.P, vector)
            pofa = lcfa.maptrans(lambda a, v: (a, poly2str(Poly(reb.topoly(v)))))
            with open(poly, "w") as f:
                print(f">>> Writing {poly} in format dot\n")
                pofa.todot(f)
        if cut is None:
            if self.r is None:
                raise ValueError("the substitution is not (X^k-)Pisot!")
            cut = compcut(lcfa, self.P, vector, lineq[1], self.verbose)
        elif self.verbose:
            print(f"Bypassing bound computation! cut={cut}")
        fa = flatten(lcfa, self.P, cut, self.verbose)
        if self.stats:
            prstats(f"Linear equation automaton {lineq}", fa)
        fa = fa.trim()
        if self.stats:
            prstats(f"Trimmed automaton {lineq}", fa)
        if self.verbose:
            obscut = realcut(fa, companion(self.P))
            print(f"Real bound is {obscut} (vs {cut}).")
        if minimize:
            fa = fa.minimize()
            if self.stats:
                prstats("After minimization", fa)
        if dorevector:
            revector(fa, lcfa, self.P)
        if not msd:
            fa = nfa.rev(fa, fa.ma.get(tuple([0] * len(lineq[0]))))
            if self.stats:
                prstats("After reversing", fa)
            fa = fa.minimize()
            if self.stats:
                prstats("After second minimization", fa)
        with open(fname, "w") as f:
            print(f">>> Writing {fname} in format {format}\n")
            if format == "Walnut":
                fa.toWalnut(f)
            elif format == "dot":
                fa.todot(f, vformat=None)
            else:
                raise ValueError(f"Invalid format name '{format}'")

    def trylinear(self, lineq, msd, minimize, vector, cut):
        if vector is None:
            vector = [0] * (self.n - 1) + [1]
        fa = lincomb(self.fa, *lineq)
        if self.stats:
            prstats("Combination automaton", fa)
        if cut is None:
            if self.r is None:
                raise ValueError("the substitution is not (X^k-)Pisot!")
            cut = compcut(fa, self.P, vector, lineq[1], self.verbose)
        elif self.verbose:
            print(f"Bypassing bound computation! cut={cut}")
        fa = flatten(fa, self.P, cut, self.verbose)
        if self.stats:
            prstats(f"Linear equation automaton {lineq}", fa)
        fa = fa.trim()
        if self.stats:
            prstats(f"Trimmed automaton {lineq}", fa)
        obscut = realcut(fa, companion(self.P))
        return obscut

    def gencheck(self, fname, msd, name):
        pre = "msd" if msd else "lsd"
        base = f"{pre}_{name}"
        with open(fname, "w") as f:
            print(f">>> Writing {fname} script\n")
            f.write(
                f"""eval add "?{base} x+y=z":
eval total "?{base} Ax Ay Ez x + y = z";
eval injL "?{base} Ax Ay Az Axx x + y = z & xx + y = z => x = xx";
eval injR "?{base} Ax Ay Az Ayy x + y = z & x + yy = z => y = yy";
eval fun "?{base} Ax Ay Az Azz x + y = z & x + y = zz => z = zz";
eval com "?{base} Ax Ay Az x + y = z <=> y + x = z";
eval distrib "?{base} Ax Ay Az (x + y) + z = x + (y + z)";
eval zeroL "?{base} Ax Ay x+0=y <=> x=y";
eval zeroR "?{base} Ax Ay 0+x=y <=> x=y";
eval succL "?{base} Ax Ay Az Eyy Ezz yy=y+1 & zz=z+1 & x+y=z <=> x+yy=x+zz";
eval succR "?{base} Ax Ay Az Eyy Ezz yy=y+1 & zz=z+1 & x+y=z <=> yy+x=zz+x";
eval basics "?{base} Ea Eaa Eaaa a = 1 & aa = 2 & aaa = 3 & a + aa = aaa & a + 4 = 5 & aa + aaa = 5 & aaa + 5 = 8";
"""
            )



import numpy as np
from math import floor


def row(i, r, n):
    if i < len(r):
        return [pow(r[i], j) for j in range(n)]
    else:
        return [1 if j == i - len(r) else 0 for j in range(n)]


def matP(P):
    n = P.degree()
    r = list(P.roots())
    r.sort(key=lambda v: abs(v), reverse=True)
    assert len(r) == n
    k = 0
    while abs(r[-1]) < 1e-4:
        k += 1
        r.pop(-1)
    return np.array([row(i, r, n) for i in range(n)]).transpose(), abs(r[0]), k


def boundit(P, u, vobj):
    k = P.degree()
    A, th, kX = matP(P)
    gammas = np.dot(np.linalg.inv(A), u[:k])
    alpha = sum([abs(gammas[i]) / (1 - abs(A[1, i])) for i in range(1, k - kX)])
    alpha += sum([abs(gammas[i]) for i in range(k - kX, k)])
    beta = alpha + abs(gammas[0]) * th / (th - 1)
    K = abs(vobj)
    return [(K * pow(th, i)) for i in range(k)], [
        (alpha + beta * pow(th, i)) for i in range(k)
    ]
    return [floor(K * pow(th, i)) for i in range(k)], [
        floor(alpha + beta * pow(th, i)) for i in range(k)
    ]



"""Bareiss algorithm for integer matrices
https://gist.github.com/vxgmichel/080e9999a1020711f27cd60b5c2d14de"""


def do_piv(a, b, c, d, e):
    x = a * d - b * c
    q, r = divmod(x, e)
    assert r == 0
    return q


def adjugate(a):
    sign = 1
    prev = piv = 1
    n = len(a)
    am = [
        list(map(int, row)) + [0] * i + [1] + [0] * (n - i - 1)
        for i, row in enumerate(a)
    ]
    for c in range(n):
        if am[c][c] == 0:
            for i in range(c + 1, n):
                if am[i][c] != 0:
                    break
            else:
                return 0, 0, None
            am[c], am[i], sign = am[i], am[c], -sign
        prev, piv = piv, am[c][c]
        for i, row in enumerate(am):
            if i == c:
                continue
            am[i] = [do_piv(piv, y, row[c], x, prev) for x, y in zip(row, am[c])]
    det = sign * piv
    return sign, det, [row[n:] for row in am]



from argparse import (
    ArgumentParser,
    BooleanOptionalAction,
    Action,
    ArgumentError,
)
from pathlib import Path as P
from .api import Study
from .argmisc import Formatter
from . import __version__


class valid_dir(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        path = P(values[0])
        if not path.is_dir():
            raise ArgumentError(self, f"{values[0]} is not an existing directory")
        setattr(namespace, self.dest, path)


class valid_vector(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            U = list(map(int, values[0].strip().split(",")))
            setattr(namespace, self.dest, U)
        except:
            raise ArgumentError(self, f"{values[0]} is not a valid vector")


class valid_equation(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            fname = P(values[1])
        except:
            raise ArgumentError(self, f"{values[1]} is not a valid filename")
        try:
            vec, cst = values[0].strip().split("=")
            b = int(cst.strip())
            aa = list(map(int, vec.strip().split(",")))
            setattr(namespace, self.dest, (aa, b, fname))
        except:
            raise ArgumentError(self, f"{values[0]} is not a valid vector")


def main():
    parser = ArgumentParser(
        prog="licofage",
        formatter_class=Formatter,
        description="Generate deterministic finite automata associated to linear equations expressed in the numeration system associated to a given substitution.",
        epilog="""You might want to try one of the following substitutions:
     - Fibonacci "a->ab, b->a" (X^2-X-1)
     - Tribonacci "a->ab, b->ca, c->a" (X^3-X^2-X-1)
     - Some X*Pisot "a->abba, b->ab" (X^2-3X)
     - Some larger example "a->ac, b->abc, c->bcc" (X^3-4X^2+4X-2)
    
    For Walnut usage, a typical invocation is:
    $ licofage -W xpisot -D ~/Walnutdir/ "a->abba, b->ab"
    """,
    )

    parser.add_argument(
        "-V", "--version", action="version", version="%(prog)s " + __version__
    )

    parser.add_argument(
        "-v", "--verbose", action="store_true", help="display more messages"
    )
    parser.add_argument(
        "-S",
        "--stats",
        action="store_true",
        help="display statistics about computations",
    )
    parser.add_argument(
        "-W",
        "--Walnut",
        nargs=1,
        metavar="NAME",
        help="generate and organize output for Walnut usage under numeration system name NAME.",
    )
    parser.add_argument(
        "-D",
        "--outdir",
        action=valid_dir,
        default=P(""),
        nargs=1,
        metavar="DIRECTORY",
        help="generate output relatively to existing directory DIRECTORY",
    )
    parser.add_argument(
        "--simplify",
        action=BooleanOptionalAction,
        default=True,
        help="simplify the substitution before study (default: True)",
    )
    parser.add_argument(
        "--minimize",
        action=BooleanOptionalAction,
        default=True,
        help="minimize the automata before output (default: True)",
    )
    parser.add_argument(
        "--revector",
        action=BooleanOptionalAction,
        default=False,
        help="try to (re)associate vectors to the minimal automaton (default: False)",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-M",
        "--MSD",
        action="store_true",
        default=True,
        help="output automata for MSD numeration (default: MSD)",
    )
    group.add_argument(
        "-L",
        "--LSD",
        action="store_false",
        dest="MSD",
        help="output automata for LSD numeration",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["dot", "Walnut"],
        default="dot",
        help="select output automata format",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-U",
        "--vector",
        nargs=1,
        action=valid_vector,
        metavar='"x1, x2, ..., xk"',
        help="use this vector to generate the bounds",
    )
    group.add_argument(
        "-B",
        "--bound",
        nargs=1,
        action=valid_vector,
        metavar='"x1, x2, ..., xk"',
        help="use this vector as manual bounds",
    )
    parser.add_argument("subst", help="substitution considered")
    parser.add_argument(
        "-d",
        "--dfao",
        nargs="?",
        metavar="filename",
        const=P("morphism.txt"),
        type=P,
        help="generate the substitution automaton with output",
    )
    parser.add_argument(
        "-P",
        "--parent",
        nargs="?",
        metavar="filename",
        const=P("parent.txt"),
        type=P,
        help="generate the parent substitution automaton with output",
    )

    parser.add_argument(
        "-n",
        "--numsys",
        nargs="?",
        metavar="filename",
        const=P("numeration.txt"),
        type=P,
        help="generate the numeration system automaton",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-a",
        "--addition",
        nargs="?",
        metavar="filename",
        const=P("addition.txt"),
        type=P,
        help='generate the additon automaton (shortcut for --linear "1, 1, -1 = 0")',
    )
    group.add_argument(
        "-l",
        "--linear",
        default=None,
        action=valid_equation,
        nargs=2,
        metavar=('"a1, a2, ..., ak = b"', "filename"),
        help="generate the automaton of the given linear equation \sum_i a_i x_i = b. (this option can be used multiple times)",
    )
    parser.add_argument(
        "-s",
        "--seq",
        nargs="?",
        metavar="filename",
        const=P("sequence.dot"),
        type=P,
        help="generate the sequence automaton of the linear equation/addition",
    )
    parser.add_argument(
        "-p",
        "--poly",
        nargs="?",
        metavar="filename",
        const=P("poly.dot"),
        type=P,
        help="generate the polynomial automaton of the linear equation/addition",
    )

    parser.add_argument(
        "-c",
        "--check",
        nargs="?",
        metavar="filename",
        const=P("check.txt"),
        type=P,
        help="generate Walnut addition verification script",
    )

    def genparentdir(fname):
        parent = fname.parent
        if not parent.is_dir():
            parent.mkdir(parents=True)
        return fname

    name = None
    args = parser.parse_args()
    if args.Walnut is not None:
        args.format = "Walnut"
        name = args.Walnut[0].lower()
        endian = "msd_" if args.MSD else "lsd_"
        args.dfao = P("Word Automata Library") / P(f"{name.title()}.txt")
        args.parent = P("Word Automata Library") / P(f"{name.title()}Parent.txt")
        args.numsys = P("Custom Bases") / P(f"{endian}{name}.txt")
        if args.linear is None:
            args.addition = P("Custom Bases") / P(f"{endian}{name}_addition.txt")
        args.check = P("Command Files") / P(f"check_{name}.txt")

    try:
        if args.check is not None and name is None:
            raise ValueError("--check only available with --Walnut!")
        stud = Study(args.subst, args.verbose, args.stats, args.simplify)
        if args.dfao is not None:
            stud.gendfao(genparentdir(args.outdir / args.dfao), args.format)
        if args.parent is not None:
            stud.genparentdfao(genparentdir(args.outdir / args.parent), args.format)
        if args.numsys is not None:
            stud.gennumsys(
                genparentdir(args.outdir / args.numsys),
                args.MSD,
                args.format,
                args.minimize,
            )
        if args.addition is not None:
            args.linear = ([1, 1, -1], 0, args.addition)
        if args.linear is not None:
            (aa, b, fname) = args.linear
            stud.genlinear(
                genparentdir(args.outdir / fname),
                (aa, b),
                args.MSD,
                args.format,
                args.minimize,
                genparentdir(args.outdir / args.seq) if args.seq is not None else None,
                genparentdir(args.outdir / args.poly)
                if args.poly is not None
                else None,
                args.vector,
                args.bound,
                args.revector,
            )
        if args.check is not None:
            stud.gencheck(genparentdir(args.outdir / args.check), args.MSD, name)
        if args.verbose:
            print("Done.")
    except ValueError as e:
        print(f"*** {str(e)}. Sorry!")
    except AssertionError as e:
        print(f"### {str(e)}. Sorry!")



from .polytools import unfrac
from .lrseq import rebaser
from .ubound import boundit
from .misc import stra
import numpy as np


def clean(l, vl, s):
    return (l, tuple(map(lambda v: tuple(list(v)), vl)), s)


def value(vl):
    res = 0
    for i, l in enumerate(vl):
        res += l[i]
    return res


def compcut(add, P, U, vobj, verbose, preverb=""):
    n = P.degree()
    boundK, boundC = boundit(unfrac(P), U, vobj)
    if verbose:
        print(f"{preverb}(P,u) vector bound: {stra(boundK)} + {stra(boundC)}*C")
    reb = rebaser(P, U)
    addpo = add.maptrans(lambda a, v: (a, tuple(reb.topoly(v))))
    mi, ma = [0] * n, [0] * n
    for s, a, v, t in addpo.alltrans():
        for i in range(n):
            if v[i] < mi[i]:
                mi[i] = v[i]
            if v[i] > ma[i]:
                ma[i] = v[i]
    C = max(abs(sum(mi)), abs(sum(ma)))
    cut = [b + C * a for a, b in zip(boundC, boundK)]
    if verbose:
        print(f"{preverb}Fast poly bound: C in [{stra(mi)}, {stra(ma)}] => {stra(cut)}")
    return cut


def realcut(res, M):
    n = len(next(res.alltrans())[2])
    obscut = [0] * n
    q0 = res.getinitial()
    seen = set()
    todo = [(q0, tuple([0] * n))]
    while todo:
        cur, curv = todo.pop()
        for i in range(n):
            obscut[i] = max(obscut[i], abs(curv[i]))
        for a, v, t in res.gettrans(cur):
            if t not in seen:
                seen.add(t)
                nxtv = tuple(list(np.dot(M, curv) + np.array(v)))
                todo.append((t, nxtv))
    return obscut



import numpy as np


def stra(l):
    return f"({', '.join(map(str,l))})"


def poly2str(P, var="X", mult="", decreasing=True):
    l = list(P)
    res = []
    for i, c in enumerate(l):
        if abs(c) < 1e-5:
            c = 0
        try:
            cs = f"{c:+g}"
        except:
            if c < 0:
                cs = f"{c}"
            else:
                cs = f"+{c}"
        if cs == "+0":
            continue
        if cs == "+1":
            if i == 0:
                res.append("+1")
            elif i == 1:
                res.append(f"+{var}")
            else:
                res.append(f"+{var}^{i}")
        elif cs == "-1":
            if i == 0:
                res.append("-1")
            elif i == 1:
                res.append(f"-{var}")
            else:
                res.append(f"-{var}^{i}")
        else:
            if i == 0:
                res.append(f"{cs}")
            elif i == 1:
                res.append(f"{cs}{mult}{var}")
            else:
                res.append(f"{cs}{mult}{var}^{i}")
    if decreasing:
        res = res[::-1]
    res = "".join(res)
    if res == "":
        res = "0"
    elif res[0] == "+":
        res = res[1:]
    return res


import re


def str2poly(s, var="X"):
    s = s.replace(" ", "")
    terms = re.findall(r"([+-]?\d*\*?" + var + r"?\^?\d*)", s)
    terms = [term for term in terms if term]
    coeffs = {}
    for term in terms:
        if "X" not in term:
            coef = int(term)
            power = 0
        else:
            parts = term.split(var)
            if parts[0] in ["", "+", "-"]:
                coef = int(parts[0] + "1")
            elif parts[0] == "-":
                coef = -1
            else:
                coef = int(parts[0].replace("*", ""))

            if "^" in parts[1]:
                power = int(parts[1][1:])
            else:
                power = 1

        if power in coeffs:
            coeffs[power] += coef
        else:
            coeffs[power] = coef
    max_power = max(coeffs.keys())
    coef_list = [0] * (max_power + 1)
    for power, coef in coeffs.items():
        coef_list[power] = coef
    return np.polynomial.Polynomial(coef_list)



class mapper:
    def __init__(self):
        self.cur = 0
        self.h = {}
        self.back = []

    def __getitem__(self, s):
        if s in self.h:
            return self.h[s]
        self.h[s] = self.cur
        self.back.append(s)
        self.cur += 1
        return self.h[s]

    def __len__(self):
        return self.cur

    def __iter__(self):
        for x in self.h:
            yield x

    def get(self, s):
        return self.h[s]

    def rev(self, i):
        return self.back[i]

    def copy(self):
        res = mapper()
        res.cur = self.cur
        res.h = self.h.copy()
        res.back = self.back.copy()
        return res


class swapper:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __getitem__(self, v):
        if v == self.x:
            return self.y
        if v == self.y:
            return self.x
        return v



#cython: language_level=3, boundscheck=False, wraparound=False
"""Misc array utility functions."""

__all__ = ["arraysorter", "indexer"]
from libc.stdlib cimport qsort, malloc, free
from libc.string cimport memcpy
from cpython cimport array
import array
import cython

cdef array.array iat = array.array('i', [])

cdef int **cmpdata
cdef int cmpn
cdef int cmp_fun(const void *i, const void *j) noexcept nogil:
    cdef int x = (<int*>i)[0]
    cdef int y = (<int*>j)[0]
    cdef int k = 0
    while k < cmpn and cmpdata[k][x] == cmpdata[k][y]:
        k += 1
    if k == cmpn:
        return 0
    if cmpdata[k][x] < cmpdata[k][y]:
        return -1
    return 1

@cython.binding(True)
def arraysorter(list arraylist):
    "Sort a list of int arrays encoding tuples according to the array values in order."
    global cmpdata, cmpn
    cdef int k = len(arraylist)
    cdef int n = len(arraylist[0])
    cmpn = k
    cdef int[:] ref
    cdef int[:] idx = array.clone(iat, n, zero=False)
    cdef int[:] tmp = array.clone(iat, n, zero=False)
    cdef int *cur
    cdef int i
    cdef int j
    cmpdata = <int **>malloc(k * sizeof(int *))
    for i in range(k):
        ref = arraylist[i]
        cmpdata[i] = &ref[0]
    for i in range(n):
        idx[i] = i
    qsort(&idx[0], n, sizeof(int), &cmp_fun)
    for i in range(k):
        cur = cmpdata[i]
        for j in range(n):
            tmp[j] = cur[idx[j]]
        memcpy(&cur[0], &tmp[0], n*sizeof(int))
    free(cmpdata)


@cython.binding(True)
def indexer(array.array a, int n):
    "Compute first index in array [a] for values [0] to [n]."
    cdef array.array resa = array.clone(iat, n, zero=False)
    cdef int[:]  res = resa
    cdef int cur = -1
    cdef int last = 0
    cdef int m = len(a)
    cdef int[:] data = a
    cdef int i
    cdef int j
    for i in range(m):
        last = cur
        cur = data[i]
        if cur == last:
            continue
        for j in range(last+1, cur):
            res[j] = -1
        res[cur] = i
    return resa



from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext

ext_modules = [
    Extension(f"licofage.{pyx}", [f"src/pyx/{pyx}.pyx"])
    for pyx in "addcut dfamin arraysorter".split()
]


class LazyImportBuildExtCmd(build_ext):
    def run(self):
        import numpy as np

        self.include_dirs.append(np.get_include())
        super().run()


setup(
    cmdclass={"build_ext": LazyImportBuildExtCmd},
    ext_modules=ext_modules,
)
